/*
    JV Library
    Copyright (C) 2014  Jouko Juhani Virtanen

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef _ConvertStructures_included_
#define _ConvertStructures_included_

# include "AtomIDs.h"
# include "Constants.h"
# include "VectorManip.h"
# include "MinMax.h"
# include "Structures.h"

void ConvertCubeToAtom(CubeStruct &Cube, AtomStruct &Atom)
{
	Atom.AtomName="O";
	Atom.ResidueName="HOH";
	Atom.HetAtom=true;
	Atom.x=Cube.x;
	Atom.y=Cube.y;
	Atom.z=Cube.z;
	Atom.weight=Cube.density;
	Atom.BFactor=Cube.density;
        Atom.atomid=CUBE;
	Atom.AtomNumber=Cube.IntDist;
	Atom.ResidueNum=Cube.AtomType;
        if (Cube.concave) Atom.Occupancy=1.0;
        else Atom.Occupancy=0.0;
}

void ConvertCubesToAtoms(vector<CubeStruct> &Cubes, vector<AtomStruct> &Atoms)
{
	int i, ncubes=Cubes.size();
	AtomStruct Atom;
	Atom.AtomName="O";
	Atom.ResidueName="HOH";
	Atom.HetAtom=true;
	DeleteVector(Atoms);
	for (i=0;i<ncubes;i++)
	{
		Atom.x=Cubes[i].x;
		Atom.y=Cubes[i].y;
		Atom.z=Cubes[i].z;
		Atom.weight=Cubes[i].density;
		Atom.atomid=CUBE;
		Atom.AtomNumber=Cubes[i].IntDist;
		Atom.ResidueNum=Cubes[i].AtomType;
                SafePushBack(Atoms, Atom, "Atoms in ConvertCubesToAtoms");
	}
}

void GetBins(Real xgrid, Real ygrid, Real zgrid, Real xmin, Real ymin, Real zmin, int &xbin, int &ybin, int &zbin, Real CubeSize)
{
        xbin=int((xgrid-xmin)/CubeSize+0.5);
        ybin=int((ygrid-ymin)/CubeSize+0.5);
        zbin=int((zgrid-zmin)/CubeSize+0.5);
}

void GetBins(CubeStruct &cube, Real xmin, Real ymin, Real zmin, int &xbin, int &ybin, int &zbin, Real CubeSize)
{
        GetBins(cube.x, cube.y, cube.z, xmin, ymin, zmin, xbin, ybin, zbin, CubeSize);
}

void LatticeDimensions(vector<CubeStruct> &cubes, Real &xmin, Real &ymin, Real &zmin, int &MaxXBin, int &MaxYBin, int &MaxZBin, Real CubeSize)
{
	Real xmax, ymax, zmax;
	cout <<"In LatticeDimensions"<<endl;
	MinMax(xmin, ymin, zmin, xmax, ymax, zmax, cubes);
	GetBins(xmax, ymax, zmax, xmin, ymin, zmin, MaxXBin, MaxYBin, MaxZBin, CubeSize);
	MaxXBin++;
	MaxYBin++;
	MaxZBin++;
}

void LatticeDimensions(Real xmin, Real ymin, Real zmin, Real xmax, Real ymax, Real zmax, int &MaxXBin, int &MaxYBin, int &MaxZBin, Real CubeSize)
{
	GetBins(xmax, ymax, zmax, xmin, ymin, zmin, MaxXBin, MaxYBin, MaxZBin, CubeSize);
        MaxXBin++;
        MaxYBin++;
        MaxZBin++;
}

void GetLatticeDimensions(lattice &Cubes, int &MaxXBin, int &MaxYBin, int &MaxZBin)
{
        MaxXBin=Cubes.size();
        if (MaxXBin!=0)
        {
                MaxYBin=Cubes[0].size();
                if (MaxYBin!=0)
                {
                        MaxZBin=Cubes[0][0].size();
                }
                else
                {
                        cout <<"ERROR: Lattice dimension=0"<<endl;
                        exit(EXIT_FAILURE);
                }
       }
       else
       {
                cout <<"ERROR: Lattice dimensions=0"<<endl;
                exit(EXIT_FAILURE);
       }
}


void InitializeLattice(lattice &CubeLattice, int MaxXBin, int MaxYBin, int MaxZBin)
{
	CubeStruct cube;
	vector<CubeStruct> CubeVector;
	vector< vector<CubeStruct> > CubeGrid;
	cout <<"In InitializeLattice"<<endl;
	InitializeCube(cube);
	SafeAlloc(CubeVector, cube, MaxZBin, "CubeVector");
	cout <<"Allocated CubeVector"<<endl;
        SafeAlloc(CubeGrid, CubeVector, MaxYBin, "CubeGrid");
        cout <<"Allocated CubeGrid"<<endl;
	SafeAlloc(CubeLattice, CubeGrid, MaxXBin, "CubeLattice");
	cout <<"Initialized lattice"<<endl;
}

void SetCoordinates(lattice &CubeLattice, Real xmin, Real ymin, Real zmin, Real CubeSize)
{
	int MaxXBin, MaxYBin, MaxZBin;
	Real xgrid, ygrid, zgrid;
	MaxXBin=CubeLattice.size();
	MaxYBin=CubeLattice[0].size();
	MaxZBin=CubeLattice[0][0].size();
	xgrid=xmin-CubeSize;
	cout <<"In SetCoordinates"<<endl;;
	for (int j=0;j<MaxXBin;j++)
	{
                //cout <<"MaxXBin= "<<MaxXBin<<" MaxYBin= "<<MaxYBin<<" MaxZBin= "<<MaxZBin<<endl;
		xgrid+=CubeSize;
		ygrid=ymin-CubeSize;
		for (int k=0;k<MaxYBin;k++)
		{
			//cout <<"MaxXBin= "<<MaxXBin<<" MaxYBin= "<<MaxYBin<<" MaxZBin= "<<MaxZBin<<endl;
			ygrid+=CubeSize;
			zgrid=zmin-CubeSize;
			for (int l=0;l<MaxZBin;l++)
			{
				//cout <<"MaxXBin= "<<MaxXBin<<" MaxYBin= "<<MaxYBin<<" MaxZBin= "<<MaxZBin<<endl;
				//cout <<"j= "<<j<<" k= "<<k<<" l= "<<l<<endl;
				zgrid+=CubeSize;
				CubeLattice[j][k][l].x=xgrid;
				CubeLattice[j][k][l].y=ygrid;
				CubeLattice[j][k][l].z=zgrid;
                                CubeLattice[j][k][l].LocatedIn=InSolution;
			}
		}
	}
        cout <<"Leaving SetCoordinates"<<endl;
}

void ConvertCubeVectorToLattice(vector<CubeStruct> &cubes, lattice &CubeLattice, Real CubeSize)
{
	int NumCubes=cubes.size();
	int MaxXBin, MaxYBin, MaxZBin;
	int xbin, ybin, zbin;
	Real xmin, ymin, zmin;
	cout <<"In ConvertCubeVerctorToLattice"<<endl;
	LatticeDimensions(cubes, xmin, ymin, zmin, MaxXBin, MaxYBin, MaxZBin, CubeSize);
	InitializeLattice(CubeLattice, MaxXBin, MaxYBin, MaxZBin);
	cout <<"xmin= "<<xmin<<" ymin= "<<ymin<<" zmin= "<<zmin<<endl;
	SetCoordinates(CubeLattice, xmin, ymin, zmin, CubeSize);
	cout <<"After SetCoordinates.  About to enter Print3DVectorSize"<<endl;
        print3DVectorSize(CubeLattice, "CubeLattice");
	cout <<"MaxXBin= "<<MaxXBin<<" MaxYBin= "<<MaxYBin<<" MaxZBin= "<<MaxZBin<<endl;
	//CubesToPdb(cubes, "/home/jouko/project/WAXS/test/pdb/cubes.pdb");
	for (int n=NumCubes-1;n>=0;n--)
	{
		GetBins(cubes[n], xmin, ymin, zmin, xbin, ybin, zbin, CubeSize);
		//PrintCubeInfo(cubes[n]);
		//cout <<"xbin= "<<xbin<<" ybin= "<<ybin<<" zbin= "<<zbin<<endl;
		//cout <<endl;
		CubeLattice[xbin][ybin][zbin]=cubes[n];
		//cubes.pop_back();
	}
	for (int j=0;j<MaxXBin;j++)
	{
		for (int k=0;k<MaxYBin;k++)
		{
			for (int l=0;l<MaxZBin;l++)
			{
                                //cout <<"j= "<<j<<" MaxXBin= "<<MaxXBin<<endl;
                                //cout <<"k= "<<k<<" MaxYBin= "<<MaxYBin<<endl;
                                //cout <<"l= "<<l<<" MaxZBin= "<<MaxZBin<<endl;
				if (CubeLattice[j][k][l].x==0 && CubeLattice[j][k][l].y==0 && CubeLattice[j][k][l].z==0)
				{
					cout <<"j= "<<j<<" k= "<<k<<" l= "<<l<<endl;
					PrintCubeInfo(CubeLattice[j][k][l]);
				}
			}
		}
	}
	//DensityMapToPdb(CubeLattice, "/home/jouko/project/WAXS/test/pdb/ConvertCubeVectorToLattice.pdb");
}

void ConvertAtomsToCubes(vector<AtomStruct> &Atoms, vector<CubeStruct> &Cubes)
{
	int i, natom=Atoms.size();
	CubeStruct Cube;
        InitializeCube(Cube);
        SafeAlloc(Cubes, Cube, natom, "Cubes in Convert AtomsToCubes");
	for (i=0;i<natom;i++)
	{
		Cubes[i].x=Atoms[i].x;
		Cubes[i].y=Atoms[i].y;
		Cubes[i].z=Atoms[i].z;
		Cubes[i].density=Atoms[i].BFactor;
		Cubes[i].AtomType=Atoms[i].ResidueNum;
		Cubes[i].IntDist=Atoms[i].AtomNumber;
	}
}

void readPdb(string PdbFile, vector<CubeStruct> &Cubes)
{
        ProteinStruct Protein;
        readPdb(PdbFile, Protein);
        removeNonHetatm(Protein.Atoms);
        ConvertAtomsToCubes(Protein.Atoms, Cubes);
}

void readPdb(string PdbFile, lattice &Cubes, Real CubeSize)
{
        vector<AtomStruct> Atoms;
        vector<CubeStruct> CubeVector;
        readPdb(PdbFile, Atoms);
        removeNonHetatm(Atoms);
        ConvertAtomsToCubes(Atoms, CubeVector);
        ConvertCubeVectorToLattice(CubeVector, Cubes, CubeSize);
}

int CountNonZeroCubes(lattice &CubeLattice)
{
	int ncube=0;
	int MaxXBin, MaxYBin, MaxZBin;

	MaxXBin=CubeLattice.size();
	MaxYBin=CubeLattice[0].size();
	MaxZBin=CubeLattice[0][0].size();

	for (int j=0;j<MaxXBin;j++)
	{
		for (int k=0;k<MaxYBin;k++)
		{
			for (int l=0;l<MaxZBin;l++)
			{
				if (CubeLattice[j][k][l].density!=0) ncube++;
			}
		}
	}
	return ncube;
}

void ConvertCubeLatticeToCubeVector(lattice &CubeLattice, vector<CubeStruct> &cubes)
{
	int ncube=0;
	int MaxXBin, MaxYBin, MaxZBin;
	CubeStruct cube;

	InitializeCube(cube);
	MaxXBin=CubeLattice.size();
	MaxYBin=CubeLattice[0].size();
	MaxZBin=CubeLattice[0][0].size();
	cubes.clear();
	SafeAlloc(cubes, cube, MaxXBin*MaxYBin*MaxZBin, "cubes");
	cout <<"In ConvertCubeLatticeToCubeVector"<<endl;
	cout <<"MaxXBin= "<<MaxXBin<<" MaxYBin= "<<MaxYBin<<" MaxZBin= "<<MaxZBin<<endl;
	for (int j=0;j<MaxXBin;j++)
	{
		//cout <<"MaxXBin= "<<MaxXBin<<" j= "<<j<<endl;
		for (int k=0;k<MaxYBin;k++)
		{
			for (int l=0;l<MaxZBin;l++)
			{
				if (CubeLattice[j][k][l].density!=InSolution)
				{
					cubes[ncube]=CubeLattice[j][k][l];
					//cout <<"ncube= "<<ncube<<" NonZeroCubes= "<<NonZeroCubes<<endl;
					ncube++;
				}
			}
		}
	}
	cout <<"At end of ConvertCubeLatticeToCubeVector"<<endl;
}

#endif
