/*
    JV Library
    Copyright (C) 2014  Jouko Juhani Virtanen

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

//There is an analytical way to calculate RMSD, but this uses steepest descent.
# include <vector>
# include <iostream>

using namespace std;

# include <numeric>

# include "AtomUtils.h"
# include "Constants.h"
# include "Structures.h"
# include "TypeDef.h"
# include "VectorManip.h"
# include "ReadPdb.h"
# include "PrintPdb.h"
# include "MathUtils.h"

Real RMSDNoAlign(vector<AtomStruct> &Atoms1, vector<AtomStruct> &Atoms2)
{
        Real dist, rmsd=0, NumAve=0;
        int natom=Atoms1.size();
        //if (natom!=Atoms2.size()) 
        //{
        //	cout <<"Error the two molecules do not have the same number of atoms.";
        //	exit(0);
        //}
        for (int i=0;i<natom;i++)
        {
                if (Atoms1[i].AtomName=="CA" || Atoms1[i].AtomName=="O1P")
                {
                        dist=atomSquareDistance(Atoms1[i], Atoms2[i]);
                        rmsd+=dist;
                        NumAve+=1.0;
                }
        }
        rmsd/=NumAve;
        rmsd=sqrt(rmsd);
	if (isnan(rmsd))
	{
		string errorStr="RMSD is nan. NumAve= "+RealToStr(NumAve);
		error(errorStr, __LINE__, __FILE__);
	}
        return rmsd;
}

Real RMSDNoAlign(vector<AtomStruct> &Atoms1, vector<AtomStruct> &Atoms2, vector<bool> &use)
{
        Real dist, rmsd=0, NumAve=0;
        int natom1=Atoms1.size();
        int natom2=Atoms2.size();
        natom1=Atoms1.size();
        //if (natom!=Atoms2.size()) 
        //{
        //	cout <<"Error the two molecules do not have the same number of atoms.";
        //	exit(0);
        //}
        for (int i=0;i<natom1;i++)
        {
                if (use[i])
                {
			if (i>=natom2)
			{
				string errorStr="Pdbs do not match";
				errorStr+=" natom1= "+toStr(natom1);
				errorStr+=" natom2= "+toStr(natom2);
				error(errorStr, __LINE__, __FILE__);
			}
                        dist=atomSquareDistance(Atoms1[i], Atoms2[i]);
                        rmsd+=dist;
                        NumAve+=1.0;
                }
        }
        rmsd/=NumAve;
        rmsd=sqrt(rmsd);
	if (isnan(rmsd))
	{
		cout <<"NumAve= "<<NumAve<<endl;
		for (int i=0;i<natom1;i++)
		{
			printAtomInfo(Atoms1[i]);
		}
		string errorStr="RMSD is nan. NumAve= "+RealToStr(NumAve);
		error(errorStr, __LINE__, __FILE__);
	}
        return rmsd;
}

Real RMSDNoAlignAll(vector<AtomStruct> &Atoms1, vector<AtomStruct> &Atoms2)
{
        Real dist, rmsd=0, NumAve=0;
        int i, natom;
        natom=Atoms1.size();
        for (i=0;i<natom;i++)
        {
                //cout <<endl;
                //printAtomInfo(Atoms1[i]);
                //printAtomInfo(Atoms2[i]);
                //cout <<endl;
                dist=atomSquareDistance(Atoms1[i], Atoms2[i]);
                rmsd+=dist;
                NumAve+=1.0;
        }
        rmsd/=NumAve;
        rmsd=sqrt(rmsd);
        return rmsd;
}

Real EvaluateOrientation(vector<AtomStruct> Atoms1, vector<AtomStruct> Atoms2, Real theta, Real phi, Real psi, vector<bool> &use)
{
        rotateAtomsAroundVector(Atoms2, use, theta, phi, psi);
        return RMSDNoAlign(Atoms1, Atoms2, use);
}
/*
void Rotate(vector<AtomStruct> &Atoms, Real step, int RotationType)
{
        if (RotationType==X) rotateAtoms(Atoms, step, 0, 0);
        else if (RotationType==Y) rotateAtoms(Atoms, 0, step, 0);
        else if (RotationType==Z) rotateAtoms(Atoms, 0, 0, step);
        else 
        {
                cout <<"Error unallowed rotation type "<<RotationType<<endl;
                exit(EXIT_FAILURE);
        }
}
*/
Real CalcDerivative(vector<AtomStruct> Atoms1, vector<AtomStruct> Atoms2, Real step, int RotationType, Real theta, Real phi, Real psi, vector<bool> &use)
{
        Real derivative, dRMSD, rmsd, rmsd_old;
        rmsd_old=EvaluateOrientation(Atoms1, Atoms2, theta, phi, psi, use);
        if (RotationType==X) theta+=step;
        else if (RotationType==Y) phi+=step;
        else if (RotationType==Z) psi+=step;
        rmsd=EvaluateOrientation(Atoms1, Atoms2, theta, phi, psi, use);
        if (RotationType==X) theta-=step;
        else if (RotationType==Y) phi-=step;
        else if (RotationType==Z) psi-=step;
        dRMSD=rmsd-rmsd_old;
        //cout <<"RotationType= "<<RotationType<<" rmsd= "<<rmsd<<" rmsd_old= "<<rmsd_old<<" dRMSD= "<<dRMSD<<endl;
        //cout <<"step= "<<step<<endl;
        //cout <<endl;
        if (dRMSD>0)
        {
                step*=-0.5;
                derivative=CalcDerivative(Atoms1, Atoms2, step, RotationType, theta, phi, psi, use);		
        }
        else derivative=dRMSD/step;
        return derivative;
}

Real CalcGradient(vector<AtomStruct> Atoms1, vector<AtomStruct> Atoms2, Real derivative[3], Real theta, Real phi, Real psi, vector<bool> &use)
{
        int i;
        Real gradient=0, step=0.0001;
        for (i=0;i<3;i++) derivative[i]=CalcDerivative(Atoms1, Atoms2, step, i, theta, phi, psi, use);
        for (i=0;i<3;i++) gradient+=derivative[i]*derivative[i];
        return gradient;
}

Real GridSearch(vector<AtomStruct> Atoms1, vector<AtomStruct> Atoms2, Real &theta, Real &phi, Real &psi, vector<bool> &use)
{
        Real BestTheta, BestPhi, BestPsi;
        Real BestRmsd, rmsd;
        Real ThetaInc, PhiInc, PsiInc;
        ThetaInc=2.0*pi/5.0;
        PhiInc=2.0*pi/5.0;
        PsiInc=2.0*pi/5.0;
        BestTheta=theta; BestPhi=phi; BestPsi=psi;
        BestRmsd=EvaluateOrientation(Atoms1, Atoms2, theta, phi, psi, use);
        for (theta=0;theta<2.0*pi;theta+=ThetaInc)
        {
                for (phi=0;phi<2.0*pi;phi+=PhiInc)
                {
                        for (psi=0;psi<2.0*pi;psi+=PsiInc)
                        {
                                rmsd=EvaluateOrientation(Atoms1, Atoms2, theta, phi, psi, use);
                                //cout <<"rmsd= "<<rmsd<<" psi= "<<psi<<endl;
                                if (rmsd<BestRmsd)
                                {
                                        BestRmsd=rmsd; BestTheta=theta; BestPhi=phi; BestPsi=psi;
                                        //cout <<"rmsd= "<<rmsd<<" theta= "<<theta<<" phi= "<<phi<<" psi= "<<psi<<endl;
                                }
                        }
                }
        }
        theta=BestTheta; phi=BestPhi, psi=BestPsi;
        return BestRmsd;
}

void GetStep(Real derivative[], Real step[], Real TotalStep)
{
        int i;
        Real TotalSqr=0;
        for (i=0;i<3;i++) TotalSqr+=derivative[i]*derivative[i];
        for (i=0;i<3;i++) step[i]=derivative[i]*TotalStep/sqrt(TotalSqr);
}

Real MinimizeAlongVector(vector<AtomStruct> Atoms1, vector<AtomStruct> Atoms2, Real derivative[], Real &theta, Real &phi, Real &psi, vector<bool> &use)
{
        int i, nstep=0, maxstep=10000;
        Real BestTheta, BestPhi, BestPsi;
        Real dRMSD, rmsd, rmsd_old, TotalStep=0.001;
        Real dRMSDCriteria, GradientCriteria, TotalStepSqr, SqrGradient;
        Real step[3];
        GradientCriteria=1.0e-15; dRMSDCriteria=1.0e-15;
        BestTheta=theta; BestPhi=phi; BestPsi=psi;
        rmsd_old=EvaluateOrientation(Atoms1, Atoms2, theta, phi, psi, use);
        GetStep(derivative, step, TotalStep);
        while (true)
        {
                theta+=step[X];
                phi+=step[Y];
                psi+=step[Z];
                rmsd=EvaluateOrientation(Atoms1, Atoms2, theta, phi, psi, use);
                dRMSD=rmsd-rmsd_old;
                if (rmsd>rmsd_old)
                {
                        theta-=step[X];
                        phi-=step[Y];
                        psi-=step[Z];
                        for (i=0;i<3;i++) step[i]=-step[i]*0.5;
                        theta=BestTheta; phi=BestPhi; psi=BestPsi;
                }
                else
                {
                        rmsd_old=rmsd;
                        for (i=0;i<3;i++) step[i]*=1.2;
                        BestTheta=theta; BestPhi=phi; BestPsi=psi;
                }
                TotalStepSqr=0;
                for (i=0;i<3;i++) TotalStepSqr+=step[i]*step[i];
                SqrGradient=dRMSD*dRMSD/TotalStepSqr;
                if (SqrGradient<GradientCriteria || abs(dRMSD)<dRMSDCriteria) break;
		if (nstep>maxstep) break;
		nstep++;
        }
        theta=BestTheta; phi=BestPhi; psi=BestPsi;
        rmsd=EvaluateOrientation(Atoms1, Atoms2, BestTheta, BestPhi, BestPsi, use);
        return rmsd;
}

Real gradientMinimization(vector<AtomStruct> &Atoms1, vector<AtomStruct> &Atoms2, Real &theta, Real &phi, Real &psi, vector<bool> &use)
{
        Real dRMSD, rmsd=0, rmsd_old, ConvergenceCriterion;
        Real gradient, GradientCriterion=10e-15;
        Real derivative[3];
        
	ConvergenceCriterion=10e-10;
        for (int i=0;i<3;i++) derivative[i]=0;
        rmsd_old=RMSDNoAlign(Atoms1, Atoms2, use);
        while (true)
        {
                gradient=CalcGradient(Atoms1, Atoms2, derivative, theta, phi, psi, use);
                if (gradient<GradientCriterion) break;
                rmsd=MinimizeAlongVector(Atoms1, Atoms2, derivative, theta, phi, psi, use);
                dRMSD=rmsd-rmsd_old;
                //cout <<"rmsd= "<<rmsd<<" rmsd_old= "<<rmsd_old<<endl;
                if (-dRMSD<ConvergenceCriterion) break;
                rmsd_old=rmsd;
        }
	return rmsd;
} 

Real RMSD(vector<AtomStruct> &Atoms1, vector<AtomStruct> &Atoms2, Real &theta, Real &phi, Real &psi, vector<bool> &use)
{
        Real rmsd;
        Real XOrigin, YOrigin, ZOrigin;
        Real XOrigin2, YOrigin2, ZOrigin2;
        vector<AtomStruct> CopyAtoms1, CopyAtoms2;
        //CopyVector(Atoms1, CopyAtoms1);
        //CopyVector(Atoms2, CopyAtoms2);
	CopyAtoms1=Atoms1;
	CopyAtoms2=Atoms2;
        removeWaters(CopyAtoms1);
        removeWaters(CopyAtoms2);
        calcCenter(CopyAtoms1, use, XOrigin, YOrigin, ZOrigin);
        center(CopyAtoms1, use);
        center(CopyAtoms2, use);
        GridSearch(CopyAtoms1, CopyAtoms2, theta, phi, psi, use);
	rmsd=gradientMinimization(CopyAtoms1, CopyAtoms2, theta, phi, psi, use);
	cout <<"theta= "<<theta<<" phi= "<<phi<<" psi= "<<psi<<endl;
        center(Atoms2, use);
        calcCenter(Atoms2, use, XOrigin2, YOrigin2, ZOrigin2);
        rotateAtomsAroundVector(Atoms2, use, theta, phi, psi);
        calcCenter(Atoms2, use, XOrigin2, YOrigin2, ZOrigin2);
        moveAtoms(Atoms2, XOrigin, YOrigin, ZOrigin);
        cout <<"rmsd= "<<rmsd<<endl;
        return rmsd;
}

void setDefaultUse(vector<AtomStruct> &Atoms, vector<bool> &use)
{
	int natom=Atoms.size();

	SafeAlloc(use, false, natom, "use");

	for (int i=0;i<natom;i++)
	{
		if (Atoms[i].AtomName=="CA" || Atoms[i].AtomName=="OP1") use[i]=true;
	}
}

Real RMSD(vector<AtomStruct> &Atoms1, vector<AtomStruct> &Atoms2)
{
	vector<bool> use;
	Real theta, phi, psi;

	theta=phi=psi=0;
	setDefaultUse(Atoms1, use);
	return RMSD(Atoms1, Atoms2, theta, phi, psi, use);
}

Real RMSD_CA(vector<AtomStruct> &Atoms1, vector<AtomStruct> &Atoms2)
{
	vector<bool> useCA, use1, use2;
	int natom1, natom2;
	Real rmsd;
	Real theta=0, phi=0, psi=0;
	Real XOrigin1, YOrigin1, ZOrigin1;
	Real XOrigin2, YOrigin2, ZOrigin2;
	Real dx, dy, dz;
	vector<AtomStruct> AtomsCA1, AtomsCA2;
	AtomsCA1=Atoms1;
	AtomsCA2=Atoms2;

        removeNonCA(AtomsCA1);
        removeNonCA(AtomsCA2);
	natom1=AtomsCA1.size();
	natom2=AtomsCA2.size();
	
	if (natom1!=natom2 || natom1==0)
	{
		int min;
		string errorStr="natom1= "+toStr(natom1);
		errorStr+=" natom2= "+toStr(natom2);
		min=getMin(natom1, natom2);
		for (int i=0;i<min;i++)
		{
			printAtomInfo(AtomsCA1[i]);
			printAtomInfo(AtomsCA2[i]);
			cout <<endl;
		}
		error(errorStr, __LINE__, __FILE__);
	}
	setDefaultUse(Atoms1, use1);
	setDefaultUse(Atoms2, use2);
	setDefaultUse(AtomsCA1, useCA);
        rmsd=RMSD(AtomsCA1, AtomsCA2, theta, phi, psi, useCA);
        rotateAtomsAroundVector(Atoms2, use2, theta, phi, psi);
        calcCenter(Atoms1, use1, XOrigin1, YOrigin1, ZOrigin1);
        calcCenter(Atoms2, use2, XOrigin2, YOrigin2, ZOrigin2);
	dx=XOrigin1-XOrigin2;
	dy=YOrigin1-YOrigin2;
	dz=ZOrigin1-ZOrigin2;
        moveAtoms(Atoms2, dx, dy, dz);
	return rmsd;
}

Real RMSD_CA(string pdbFile, ProteinStruct Protein2)
{
	ProteinStruct Protein1;

	readPdb(pdbFile, Protein1);
	removeDuplicateAtoms(Protein1.Atoms);
	removeNonCA(Protein1.Atoms);
	removeNonCA(Protein2.Atoms);
	return RMSD(Protein1.Atoms, Protein2.Atoms);
}

bool removeOutliers(vector<AtomStruct> &Atoms1, vector<AtomStruct> &Atoms2, Real cutOff, Real rmsd, vector<bool> &use)
{
	bool foundOutlier=false;
	int natom1, natom2, worst=0;
	Real dist, cutOff2=cutOff*cutOff;
	//Real rmsd2=rmsd*rmsd;
	Real max;

	natom1=Atoms1.size();
	natom2=Atoms2.size();

	if (natom1!=natom2 || natom1==0)
	{
		cout <<"ERROR: natom1= "<<natom1<<" natom2= "<<natom2<<endl;
		exit(EXIT_FAILURE);
	}	
	max=0;
	for (int i=0;i<natom1;i++)
	{
		if (Atoms1[i].AtomName=="CA" || Atoms1[i].AtomName=="OP1")
		{
			dist=atomSquareDistance(Atoms1[i], Atoms2[i]);
			if (dist<cutOff2) 
			{
				if (!use[i]) 
				{
					foundOutlier=true;
				}
				use[i]=true;
			}
			else 
			{
				if (use[i]) 
				{
					foundOutlier=true; 
					if (dist>max)
					{
						max=dist;
						worst=i;
					}
				}
			}
/*
			if (dist>rmsd2 && rmsd>cutOff)
			{
				if (use[i])
				{
					foundOutlier=true;
					use[i]=false;
				}
			}
*/	
		}
	}
	cout <<"max= "<<max<<" worst= "<<worst<<endl;
	if (max>0) use[worst]=false;
	return foundOutlier;
}

Real RMSD_NoOutliers(vector<AtomStruct> &Atoms1, vector<AtomStruct> &Atoms2)
{
	bool foundOutliers=true;
	vector<bool> use;
	Real theta, phi, psi;
	Real rmsd, cutOff;

	theta=phi=psi=0;	
	setDefaultUse(Atoms1, use);
	center(Atoms1, use);
        center(Atoms2, use);
	while (true)
	{
		rmsd=RMSD(Atoms1, Atoms2, theta, phi, psi, use);
		cutOff=rmsd*2.0;
		foundOutliers=removeOutliers(Atoms1, Atoms2, cutOff, rmsd, use);
		if (!foundOutliers) break;
	}
	
	return rmsd;
}

Real RMSD_NoOutliers(vector<AtomStruct> &Atoms1, vector<AtomStruct> &Atoms2, Real cutOff)
{
	bool foundOutliers=true;
	vector<bool> use;
	Real theta, phi, psi;
	Real rmsd;
	Real xave1, yave1, zave1;
	Real xave2, yave2, zave2;

	theta=phi=psi=0;	
	setDefaultUse(Atoms1, use);
	center(Atoms1, use);
        center(Atoms2, use);
	rmsd=RMSD(Atoms1, Atoms2, theta, phi, psi, use);
	while (true)
	{
		rmsd=RMSD(Atoms1, Atoms2, theta, phi, psi, use);
		calcCenter(Atoms1, use, xave1, yave1, zave1);
		calcCenter(Atoms2, use, xave2, yave2, zave2);
		foundOutliers=removeOutliers(Atoms1, Atoms2, cutOff, rmsd, use);
		if (!foundOutliers) break;
	}
	cout <<"RMSD= "<<rmsd<<endl;	
	return rmsd;
}

