/*
    JV Library
    Copyright (C) 2014  Jouko Juhani Virtanen

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

# include <sys/time.h>

# include "MathUtils.h"
# include "VectorManip.h"
# include "test_functions.h"
# include "minimize.h"
# include "time.h"

void PrintHessian(Matrix &Mat)
{
	int Size1, Size2;
	Size1=Mat.size();
	Size2=Mat[0].size();
	for (int i=0;i<Size1;i++)
	{
                for (int j=0;j<Size2;j++)
                {
                        cout <<setiosflags(ios::left)<<setw(10)<<setprecision(5)<<Mat[i][j]<<"\t";
		}
                cout <<endl;
	}
	cout <<endl;
}

Real calcVectorMagnitude(Vector &v)
{
        int Size=v.size();
        Real sum=0;

        for (int i=0;i<Size;i++)
        {
                sum+=v[i]*v[i];
        }
        return sqrt(sum);
}

void scaleVector(Vector &v, Real length)
{
        int Size=v.size();
        Real magnitude;
        magnitude=calcVectorMagnitude(v);
        for (int n=0;n<Size;n++)
        {
                v[n]*=length/magnitude;
        }
}

Real calcFletcherReeves(Real f2, Real f1)
{
        return f2*f2/(f1*f1);
}

void updateStep(Vector &derivative, Vector &derivative_old, Vector &step, Vector &step_old)
{
        int nstep=step.size();
        Real gamma;

        for (int i=0;i<nstep;i++)
        {
                if (derivative_old[i]!=0)
                {
                        gamma=calcFletcherReeves(derivative[i], derivative_old[i]);
                        step[i]=derivative[i]+gamma*step_old[i];
                }
                else step[i]=derivative[i];
        }
}

Vector solveHessian(Vector &gradient, Matrix &hessian)
{
        int ncoeff=gradient.size();
        Vector step;
        Matrix coefficient;

        SafeAlloc(step, ncoeff, "step");
        Safe2DAlloc(coefficient, ncoeff+1, ncoeff, "coefficient");
        for (int i=0;i<ncoeff;i++)
        {
                //cout <<"coefficient["<<i<<"]["<<ncoeff<<"]= "<<coefficient[i][ncoeff]<<endl;
                coefficient[ncoeff][i]=+gradient[i];
                for (int j=0;j<ncoeff;j++)
                {
                        coefficient[i][j]=hessian[i][j];
                }
        }
        equation(coefficient, step);
        return step;
}

Real calcAngle_temp(Vector v1, Vector v2)
{
	int Size=v1.size();
	Real dp=0, length1=0, length2=0;

	for (int i=0;i<Size;i++)
	{
		dp+=v1[i]*v2[i];
		length1+=v1[i]*v1[i];
		length2+=v2[i]*v2[i];
	}

	return acos(dp/(sqrt(length1*length2)));
}

void estimateGradient(Matrix &hessian, Vector &derivative, Vector &start, Vector &coefficient, Vector &step)
{
        int Size=derivative.size();
        Vector diff;

        SafeAlloc(diff, Size, "diff");
        step=derivative;
        for (int i=0;i<Size;i++) diff[i]=coefficient[i]-start[i];

        for (int i=0;i<Size;i++)
        {
                for (int j=0;j<Size;j++)
                {
                        step[i]+=hessian[i][j]*diff[j];
                }
        }
}

